From 1546916eb98367c27b5bfbef2814c14f38c2eb80 Mon Sep 17 00:00:00 2001
From: David Oberhollenzer <david.oberhollenzer@sigma-star.at>
Date: Sun, 22 Aug 2021 13:29:26 +0200
Subject: [PATCH] Cleanup bounds checks and initialization in dir reader

The sqfs_dir_reader_open_dir function tried to take a short-cut by
returning early if the target directory is empty. However, this left
some field unchanged from the previous directory.

If iterating over a directory and then deciding to enter a sub-directory
that happens to be empty, the directory reader will keep the settings
for the current directory. After calling sqfs_dir_reader_rewind, the
sub-directory will suddenly report the contents of the parent.

A similar check is added to the rewind function to not track back on
the meta data reader in that case.

The checks in the read function are also adjusted to use the same bounds
to determine if a directory can be read from and check if it is possible
to read an entry header before trying to do so.

Signed-off-by: David Oberhollenzer <david.oberhollenzer@sigma-star.at>
---
 lib/sqfs/dir_reader.c | 17 +++++++++++++----
 1 file changed, 13 insertions(+), 4 deletions(-)

diff --git a/lib/sqfs/dir_reader.c b/lib/sqfs/dir_reader.c
index 65962d8..f560069 100644
--- a/lib/sqfs/dir_reader.c
+++ b/lib/sqfs/dir_reader.c
@@ -145,15 +145,15 @@ int sqfs_dir_reader_open_dir(sqfs_dir_reader_t *rd,
 	rd->size = size;
 	rd->entries = 0;
 
-	if (rd->size <= sizeof(rd->hdr))
-		return 0;
-
 	block_start += rd->super->directory_table_start;
 
 	rd->dir_block_start = block_start;
 	rd->dir_offset = offset;
 	rd->start_size = size;
 
+	if (rd->size <= sizeof(rd->hdr))
+		return 0;
+
 	return sqfs_meta_reader_seek(rd->meta_dir, block_start, offset);
 }
 
@@ -164,7 +164,7 @@ int sqfs_dir_reader_read(sqfs_dir_reader_t *rd, sqfs_dir_entry_t **out)
 	int err;
 
 	if (!rd->entries) {
-		if (rd->size < sizeof(rd->hdr))
+		if (rd->size <= sizeof(rd->hdr))
 			return 1;
 
 		err = sqfs_meta_reader_read_dir_header(rd->meta_dir, &rd->hdr);
@@ -175,6 +175,12 @@ int sqfs_dir_reader_read(sqfs_dir_reader_t *rd, sqfs_dir_entry_t **out)
 		rd->entries = rd->hdr.count + 1;
 	}
 
+	if (rd->size <= sizeof(*ent)) {
+		rd->size = 0;
+		rd->entries = 0;
+		return 1;
+	}
+
 	err = sqfs_meta_reader_read_dir_ent(rd->meta_dir, &ent);
 	if (err)
 		return err;
@@ -200,6 +206,9 @@ int sqfs_dir_reader_rewind(sqfs_dir_reader_t *rd)
 	rd->size = rd->start_size;
 	rd->entries = 0;
 
+	if (rd->size <= sizeof(rd->hdr))
+		return 0;
+
 	return sqfs_meta_reader_seek(rd->meta_dir, rd->dir_block_start,
 				     rd->dir_offset);
 }
-- 
2.31.1

